from typing import List, Tuple, Type

import tiktoken
from langchain import LLMChain, PromptTemplate
from langchain.base_language import BaseLanguageModel
from langchain.chat_models import ChatOpenAI
from langchain.embeddings.base import Embeddings
from langchain.schema import Document

from chat.ask_codebase.store.qdrant import QdrantWrapper as Q
from chat.ask_codebase import DEBUG

prompt_template = """
Use the following pieces of context to answer the question at the end. If you don't know the answer, just say that you don't know, don't try to make up an answer.

Note:
- Please incorporate your reasoning and some code into the answer and ensure that it is included.
- Let's speculate on what code might be needed to answer the question more helpful.

Context (some codes or documents):
=====================
{context}
=====================

Question: {question}"""
PROMPT = PromptTemplate(
    template=prompt_template, input_variables=["context", "question"]
)


class StuffDocumentCodeQa:
    llm: BaseLanguageModel = ChatOpenAI(temperature=0.1)  # type: ignore

    def __init__(
        self,
        # source_name: str,
        # embedding_cls: Type[Embeddings],
        q: Q,
    ) -> None:
        # self._source_name = source_name
        # self._embedding_cls = embedding_cls
        self._q = q

    def run(self, question: str) -> Tuple[str, List[Document]]:
        # q = Q.reuse(self._source_name, self._embedding_cls)
        qdrant = self._q.qdrant

        potential_found_docs = qdrant.similarity_search(
            question,
            k=50,
        )

        enc = tiktoken.encoding_for_model("gpt-4")
        total_token = 0
        found_docs = []
        for found_doc in potential_found_docs:
            tokenized_text = enc.encode(found_doc.page_content)
            if total_token + len(tokenized_text) > 2000:
                break
            total_token += len(tokenized_text)
            found_docs.append(found_doc)

        llm_chain = LLMChain(llm=self.llm, prompt=PROMPT, verbose=DEBUG)
        completion = llm_chain.predict_and_parse(
            **{  # type: ignore
                "context": "==========\n".join(
                    [
                        f"""{found_doc.metadata['source']}:\n{found_doc.page_content}\n"""
                        for index, found_doc in enumerate(found_docs)
                    ]
                ),
                "question": question,
            }
        )

        return str(completion), found_docs
