import functools
import os
from pathlib import Path
from typing import Callable, Iterable, Iterator, NamedTuple, Optional

from langchain.schema import Document
from langchain.text_splitter import (
    CharacterTextSplitter,
    Language,
    RecursiveCharacterTextSplitter,
    TextSplitter,
)
from loguru import logger

from chat.ask_codebase.indexing.loader.base import Loader


class FileMetadata(NamedTuple):
    # Relative path
    filepath: str
    # Used to jump to the source file
    # (web url or local file path)
    reference: str
    # TODO: define "name" & "source" & "url" to be compatible with old OSS data
    # TODO: Remove the usage of them later
    name: str  # use `filepath` instead
    source: str  # remove it
    url: str  # use `reference` instead

    label: Optional[str] = None


class FileSource(NamedTuple):
    # absolute path to visit files or directories
    path: str
    # The root to compute relative path
    rel_root: Optional[str] = None
    # A function that maps filepath to reference
    ref_maker: Optional[Callable[[str], str]] = None
    # Only files that pass the filter will be loader
    file_filter: Optional[Callable[[str], bool]] = None


def get_file_splitter(
    filepath: Path, chunk_size: int = 5000, chunk_overlap: int = 50
) -> TextSplitter:
    splitter_lang_mapping = {
        ".py": Language.PYTHON,
        ".js": Language.JS,
        ".go": Language.GO,
        ".java": Language.JAVA,
        ".php": Language.PHP,
        ".rs": Language.RUST,
        ".rb": Language.RUBY,
        ".scala": Language.SCALA,
        ".swift": Language.SWIFT,
        ".md": Language.MARKDOWN,
    }

    ext = filepath.suffix
    splitter_lang = splitter_lang_mapping.get(ext, None)

    splitter: TextSplitter
    if splitter_lang:
        splitter = RecursiveCharacterTextSplitter.from_language(
            language=splitter_lang,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
        )
    else:
        splitter = CharacterTextSplitter.from_tiktoken_encoder(
            model_name="text-embedding-ada-002",
            disallowed_special=(),
            # reduce chunk size when splitting by token
            chunk_size=chunk_size / 10,
            chunk_overlap=chunk_overlap,
        )

    return splitter


def _github_url_maker(relpath: str, repo_url: str, hexsha: str = "main") -> str:
    base_url = os.path.splitext(repo_url)[0]

    return f"{base_url}/blob/{hexsha}/{relpath}"


def _local_file_maker(relpath: str, repo_dir: str) -> str:
    return os.path.join(repo_dir, relpath)


def gen_github_reference_maker(
    repo_url: str, hexsha: str = "main"
) -> Callable[[str], str]:
    return functools.partial(
        _github_url_maker, repo_url=repo_url, hexsha=hexsha
    )


def gen_local_reference_maker(repo_dir: str) -> Callable[[str], str]:
    return functools.partial(_local_file_maker, repo_dir=repo_dir)


class FileLoader(Loader):
    """
    Load source files and split them into documents
    """

    def __init__(self, sources: Iterable[FileSource]):
        self.sources = sources

    def load(
        self,
    ) -> Iterator[Document]:
        """
        Load files to documents

        """
        for source in self.sources:
            root = Path(source.path)
            file_filter = source.file_filter
            ref_maker = source.ref_maker
            rel_root = source.rel_root

            for f in self._iter_path_or_dir(root):
                if file_filter and not file_filter(str(f)):
                    # skip files that do not pass the filter
                    logger.info(f"Skip file: {f}")
                    continue

                content = f.read_text(encoding="utf-8", errors="ignore")
                rel_path = str(f.relative_to(rel_root)) if rel_root else str(f)
                reference = ref_maker(rel_path) if ref_maker else None
                metadata = FileMetadata(
                    filepath=rel_path,
                    reference=reference,
                    # TODO: deprecate them later (after generating new OSS data)
                    name=os.path.basename(rel_path),
                    source=rel_path,
                    url=reference,
                )
                doc = Document(
                    page_content=content, metadata=metadata._asdict()
                )

                splitter = get_file_splitter(filepath=f)
                docs = splitter.split_documents([doc])
                yield from docs

    def _iter_path_or_dir(self, path: Path) -> Iterator[Path]:
        """
        Iterate files in a path
        """
        if path.is_file():
            yield path

        elif path.is_dir():
            for f in path.rglob("*"):
                if f.is_file():
                    yield f
