import csv
from pathlib import Path
from typing import Callable, Dict, Iterator, NamedTuple

from langchain.schema import Document

from chat.ask_codebase.indexing.loader.base import Loader


class FunctionSource(NamedTuple):
    # The root directory fo the repo
    repo_path: str
    # repo's cag file path
    cag_path: str
    hexsha: str
    # A function that maps filepath to url:
    # url_from_filepath(hexsha: str, filename: str, start_line: int, end_line: int) -> str
    url_from_filepath: Callable[[str, str, int, int], str]


class FunctionLoader(Loader):
    """
    Load CAG functions(produced by AE) and split them into documents
    """

    def __init__(self, source: FunctionSource):
        self._source = source

    @classmethod
    def get_function_with_comment(cls, file_path, start_line, end_line):
        """
        get code(start_line~end_line) and continuous comments before start_line
        """
        with open(file_path, "r") as file:
            all_lines = file.readlines()
            target_lines = []
            in_long_comment = False
            for line_number, line in enumerate(all_lines, start=1):
                if line_number < start_line:
                    if in_long_comment:
                        if "*/" in line.strip():
                            in_long_comment = False
                        target_lines.append(line)
                    elif line.strip().startswith("/*"):
                        in_long_comment = True
                        target_lines.append(line)
                    elif line.strip().startswith(
                        "//"
                    ) or line.strip().startswith("#"):
                        target_lines.append(line)
                    else:
                        target_lines = []
                elif line_number <= end_line:
                    target_lines.append(line)
                else:
                    break
        return "".join(target_lines)

    def load(
        self,
    ) -> Iterator[Document]:
        """
        Load cag functions from csv files to Documents

        """

        # golang function
        cag_path = Path(self._source.cag_path)
        cag_nodes = []
        with open(
            cag_path / f"{self._source.hexsha}_cag_nodes.csv", "r", newline=""
        ) as file:
            reader = csv.DictReader(file)
            for row in reader:
                if not row["file_path"].startswith("backend/"):
                    continue
                if row["is_test_func"] == "True":
                    continue
                cag_nodes.append(row)
        print(len(cag_nodes))

        # function edge
        cag_edges_from_map: Dict[str, list] = {}  # group by from_node_id
        cag_edges_to_map: Dict[str, list] = {}  # group by to_node_id
        with open(
            cag_path / f"{self._source.hexsha}_cag_edges.csv", "r", newline=""
        ) as file:
            reader = csv.DictReader(file)
            for row in reader:
                if row["to_node_id"] not in cag_edges_to_map:
                    cag_edges_to_map[row["to_node_id"]] = []
                cag_edges_to_map[row["to_node_id"]].append(row["from_node_id"])
                if row["from_node_id"] not in cag_edges_from_map:
                    cag_edges_from_map[row["from_node_id"]] = []
                cag_edges_from_map[row["from_node_id"]].append(
                    row["to_node_id"]
                )

        # ignore the functions which called by more than 10 if a node called >3 functions. Because it may not helpful
        for from_node_id in cag_edges_from_map:
            if len(cag_edges_from_map[from_node_id]) > 3:
                new_to_node_ids = []
                for to_node_id in cag_edges_from_map[from_node_id]:
                    if len(cag_edges_to_map[to_node_id]) < 10:
                        new_to_node_ids.append(to_node_id)
                cag_edges_from_map[from_node_id] = new_to_node_ids
        print(len(cag_edges_from_map))
        # for a in cag_edges_from_map:
        #     print(a, len(cag_edges_from_map[a]))
        print(len(cag_edges_to_map))
        # for a in cag_edges_to_map:
        #     print(a, len(cag_edges_to_map[a]))

        code_root = Path(self._source.repo_path)
        for cag_node in cag_nodes:
            file_path = str(code_root / cag_node["file_path"])
            code = self.get_function_with_comment(
                file_path,
                int(cag_node["start_line"]),
                int(cag_node["end_line"]),
            )
            content = f"{cag_node['file_path']}#L{cag_node['start_line']}-L{cag_node['end_line']}\n```\n{code}```"

            url = self._source.url_from_filepath(
                self._source.hexsha,
                cag_node["file_path"],
                int(cag_node["start_line"]),
                int(cag_node["end_line"]),
            )
            metadata = {
                "source": cag_node["file_path"],
                "name": cag_node["name"],
                "node_id": cag_node["node_id"],
                "from_node_ids": cag_edges_to_map.get(cag_node["node_id"], []),
                "to_node_ids": cag_edges_from_map.get(cag_node["node_id"], []),
                "len": int(cag_node["end_line"]) - int(cag_node["start_line"]),
                "url": url,
            }
            doc = Document(page_content=content, metadata=metadata)

            yield doc
