import uuid
from typing import Dict, Iterable, List, Optional, Tuple, Type

import qdrant_client
from langchain.embeddings.base import Embeddings
from langchain.schema import Document
from langchain.vectorstores import Qdrant
from loguru import logger
from qdrant_client.http import models as rest

from chat.ask_codebase.indexing.embedding import Vector
from chat.ask_codebase import QDRANT_KEY, QDRANT_URL

# from enum impor


DEFAULT_DISTANCE = rest.Distance.COSINE


class PureFilterQdrant(Qdrant):
    """
    return pure filter
    """

    def _qdrant_filter_from_dict(
        self, filter: Optional[dict]
    ) -> Optional[rest.Filter]:
        from qdrant_client.http import models as rest

        if not filter:
            return None

        return rest.Filter(**filter)


# def get_embeddings(embedding_type: str) -> Embeddings:
#     """
#     Get embeddings instance from embedding type string
#     """
#     module_name, class_name = embedding_type.rsplit(".", 1)
#     module = importlib.import_module(module_name)
#     cls = getattr(module, class_name)
#     return cls()


def get_client(mode: str = ":memory:", qdrant_url: str = QDRANT_URL, qdrant_key: str= QDRANT_KEY) -> qdrant_client.QdrantClient:
    """
    Get QdrantClient instance
    mode should be one of:
        - :memory: default
        - :server: connect to QDRANT_URL and QDRANT_KEY from env
        - path/to/file: persist to file
    """
    if mode == ":server:":
        return qdrant_client.QdrantClient(
            url=qdrant_url, prefer_grpc=True, api_key=qdrant_key
        )
    elif mode == ":memory:":
        return qdrant_client.QdrantClient(":memory:")
    else:
        return qdrant_client.QdrantClient(path=mode)


class QdrantWrapper:
    """
    NOTE:
    Always use classmethod Create() / Reuse() to get a QdrantWrapper instance
    """

    NAME_SEPARATOR = "---"  # collection name separator

    def __init__(
        self,
        source_name: str,
        embedding: Embeddings,
        client: qdrant_client.QdrantClient,
    ) -> None:
        self._ds_name = source_name
        self._embedding = embedding
        self._q: Optional[PureFilterQdrant] = None

        # if the collection is existed or not
        self._is_existed = False
        self._vector_size = 0

        self.client = client

    @property
    def source_name(self) -> str:
        return self._ds_name

    @property
    def embedding(self) -> Embeddings:
        return self._embedding

    @property
    def collection_name(self) -> str:
        embedding_name = self._embedding.__class__.__name__
        return self.get_collection_name(self._ds_name, embedding_name)

    @property
    def qdrant(self) -> Optional[PureFilterQdrant]:
        if self._q is None:
            qdrant = PureFilterQdrant(
                client=self.client,
                collection_name=self.collection_name,
                embeddings=self.embedding,
            )
            self._q = qdrant
        return self._q

    @property
    def info(self) -> Optional[rest.CollectionInfo]:
        try:
            return self.client.get_collection(self.collection_name)
        except Exception:
            return None

    @staticmethod
    def get_collection_name(source_name: str, embedding_name: str) -> str:
        return f"{source_name}{QdrantWrapper.NAME_SEPARATOR}{embedding_name}"

    @classmethod
    def create(
        cls,
        source_name: str,
        embedding_cls: Type[Embeddings],
        client: qdrant_client.QdrantClient = None,
    ):
        if client is None:
            # Use server mode by default
            client = get_client(":server:")

        # TODO: manage embedding types
        embedding = embedding_cls()
        wrapper = cls(source_name, embedding, client)

        return wrapper

    @classmethod
    def reuse(
        cls,
        source_name: str,
        embedding_cls: Type[Embeddings],
        client: qdrant_client.QdrantClient = None,
    ):
        if client is None:
            # Use server mode by default
            client = get_client(":server:")

        # TODO: manage embedding types
        embedding = embedding_cls()
        wrapper = cls(source_name, embedding, client)

        collection_name = wrapper.collection_name
        try:
            info = client.get_collection(collection_name)
        except Exception as e:
            logger.warning(f"Collection {collection_name} does not exist. {e}")
            raise ValueError(
                f"Datasource {source_name} with {embedding_cls.__name__} does not exist. "
            )

        wrapper._vector_size = info.config.params.vectors.size  # type: ignore
        wrapper._is_existed = True

        qdrant = PureFilterQdrant(
            client=wrapper.client,
            collection_name=collection_name,
            embeddings=wrapper.embedding,
        )
        wrapper._q = qdrant
        return wrapper

    def insert(
        self, data: Iterable[Tuple[Vector, Document]], batch_count: int = 64
    ):
        """
        Insert vectors into Qdrant in batch.
        batch_size: How many vectors upload per-request.
        """
        buffer: List[Tuple[Vector, Document]] = []

        for vector, doc in data:
            if not self._is_existed:
                size = len(vector)
                self._recreate_collection(size)

            buffer.append((vector, doc))

            if len(buffer) >= batch_count:
                res = self.insert_many(buffer)
                if res:
                    buffer.clear()

        while buffer:
            res = self.insert_many(buffer)
            if res:
                buffer.clear()

    def insert_many(self, data: List[Tuple[Vector, Document]]) -> bool:
        logger.info(
            f"Insert {len(data)} vectors into {self.collection_name}..."
        )
        if not data:
            return True

        if not self._is_existed:
            size = len(data[0][0])
            self._recreate_collection(size)

        vectors: List[Vector] = []
        payloads: List[Dict] = []
        ids: List[str] = []

        for v, d in data:
            vectors.append(v)
            payloads.append(
                {
                    "page_content": d.page_content,
                    "metadata": d.metadata,
                }
            )
            ids.append(uuid.uuid4().hex)

        points = rest.Batch.construct(
            ids=ids,
            vectors=vectors,
            payloads=payloads,
        )

        try:
            self.client.upsert(
                collection_name=self.collection_name,
                points=points,
            )
        except Exception as e:
            logger.error(f"Failed to insert {len(data)} vectors: {e}")
            return False

        logger.info("Finish Inserting.")
        return True

    def _recreate_collection(self, vector_size: int):
        """
        Recreate collection
        """
        client = self.client

        client.recreate_collection(
            collection_name=self.collection_name,
            vectors_config=rest.VectorParams(
                size=vector_size, distance=DEFAULT_DISTANCE
            ),
        )

        self._is_existed = True
        self._vector_size = vector_size
        logger.info(
            f"Recreated collection {self.collection_name} with vector size {vector_size}."
        )
