import json
import os
import subprocess
from pathlib import Path
from typing import Any, Optional

repo_root = Path(__file__).parent.parent.parent.resolve()


def clone_repo_if_not_exist(
    repo_url: str, repo_dir: str, target_sha: str = "HEAD"
) -> str:
    """
    Clone a repo if it does not exist, return the HEAD hash
    """
    if not os.path.exists(repo_dir):
        subprocess.run(["git", "clone", repo_url, repo_dir])

    if target_sha == "HEAD":
        subprocess.run(f"cd '{repo_dir}' && git fetch --all", shell=True)
    else:
        subprocess.run(
            f"cd '{repo_dir}' && git fetch --all && git checkout {target_sha}",
            shell=True,
        )
    proc = subprocess.Popen(
        f"cd '{repo_dir}' && git rev-parse HEAD",
        shell=True,
        stdout=subprocess.PIPE,
    )
    proc.wait()
    output = proc.stdout.read().decode().strip()

    return output


def is_source_code(file_name: str) -> bool:
    """
    Check if a given file is a source code file based on its extension.

    Args:
        file_name (str): The name of the file to check.

    Returns:
        bool: True if the file is a source code file, False otherwise.
    """
    # List of meaningful source code file extensions
    source_code_extensions = [
        ".py",  # Python
        ".java",  # Java
        ".c",  # C
        ".cpp",  # C++
        ".h",  # C header
        ".hpp",  # C++ header
        ".hh",  # C++ header
        ".js",  # JavaScript
        ".ts",  # TypeScript
        ".go",  # Go
        ".rs",  # Rust
        ".rb",  # Ruby
        ".cs",  # C#
        ".m",  # Objective-C
        ".swift",  # Swift
        ".php",  # PHP
        ".kt",  # Kotlin
        ".scala",  # Scala
        ".r",  # R
        ".pl",  # Perl
        ".lua",  # Lua
        ".groovy",  # Groovy
        ".dart",  # Dart
        ".sh",  # Bash
        ".bat",  # Batch file
        ".md",  # Markdown
        ".yaml",  # YAML
        ".yml",  # YAML
    ]

    _, extension = os.path.splitext(file_name)

    return extension in source_code_extensions


def read_json(input_filename: str) -> Optional[Any]:
    """
    Reads a JSON file and returns its content as Python data structure.

    Args:
        input_filename (str): Path to the JSON file.

    Returns:
        Python data structure resulting from the JSON file decoding,
        or None if an error occurred during the loading.
    """
    try:
        with open(input_filename, "r") as infile:
            return json.load(infile)
    except FileNotFoundError:
        print(f"File {input_filename} not found.")
    except json.JSONDecodeError:
        print(f"An error occurred while decoding {input_filename}.")
    return None


def write_to_json(data: Any, output_filename: str) -> bool:
    """
    Writes data to a JSON file.

    Args:
        data (Any): The data to be written to the JSON file.
        output_filename (str): The path to the JSON file to write to.

    Returns:
        bool: True if data was written successfully, False otherwise.
    """
    try:
        with open(output_filename, "w") as outfile:
            json.dump(data, outfile)
        return True
    except FileNotFoundError:
        print(f"File {output_filename} not found.")
    except TypeError:
        print("An error occurred while encoding data to JSON.")
    return False
